//******************************************************************************
//     Name : LinkStateSim 1.3
//  Purpose : Link State Routing Simulator with Sockets
//
//   Author : Marco Benvegn
//	          hiforce@gmx.it
//            http://web.tiscali.it/hiforce/
//
//      Dev : 23h over 5 days (1-3-4-5-6 Feb 2003) (See LinkStateSim.tim)
//******************************************************************************

#include <stdio.h>
#include <winsock.h>

#define MASTER_PORT 9999
#define BASE_PORT 10000
#define QUEUE_SIZE 5
#define FLOODING_TIMEOUT 1		// in seconds
#define MAX_LINKS FD_SETSIZE	// limit imposed by select()
#define MAX_ROUTERS 1024		// TODO: use a std::vector

char *g_inputPath="input.txt";	// default input filename
bool *g_completed;				// used to detect the end of Flooding (not part of the algorithm)

//-------------------------------------------------------------------------
// Structures
//-------------------------------------------------------------------------

typedef enum {HELLO, HACK, LSU, ACK, START, DATA} packet_type;

typedef struct
{
	packet_type		type;
	unsigned int	sourceID;
	unsigned int	numberOfLinks;
	unsigned int	neighborID[MAX_LINKS];
	unsigned int	cost[MAX_LINKS]; // weight of the link between sourceID and neighborID[i]
} Packet;

typedef struct {
	unsigned int distance;		// distance from source
	unsigned int predecessor;	// predecessor on the shortest path to the source
	bool	     permanent;		// label state
} Dijkstra_label;

//-------------------------------------------------------------------------
// Tools
//-------------------------------------------------------------------------

void error(char *string)
{
	fprintf(stderr, "ERROR: %s\n", string);
	exit(1); // Close all threads
}

// Reads network description from a text file (TAB delimited)
// numberOfRouters	Output	Number Of Routers
// LSP				Output	Pointer to a heap array of numberOfRouters LSPs
void loadNetworkDescription(unsigned int &numberOfRouters, Packet* &LSP)
{
	unsigned int i;

	// Read network description
	FILE* in;
	in=fopen(g_inputPath, "rt");
	if (in==0) error("cannot open input file");

	// read numberOfRouters
	fscanf(in, "%u", &numberOfRouters);

	// prepare HELLO packets
	LSP=new Packet[numberOfRouters];
	for (i=0 ; i<numberOfRouters ; i++)
	{
		LSP[i].type=HELLO;
		LSP[i].sourceID=i;
		LSP[i].numberOfLinks=0;
	}

	unsigned int id1, id2, cost, next;
	while (fscanf(in, "%u\t%u\t%u\n", &id1, &id2, &cost)!=EOF) // TAB delimited
	{
		printf("%u\t%u\t%u\n", id1, id2, cost);

		if (id1>=numberOfRouters || id2>=numberOfRouters) error("router's ID must be < number of routers"); // TODO: relax this rule
		for (i=0 ; i<LSP[id1].numberOfLinks ; i++) if (LSP[id1].neighborID[i]==id2) error("duplicate link");

		next=LSP[id1].numberOfLinks++;
		if (next>=MAX_LINKS) error("max number of links per router exceeded");
		LSP[id1].neighborID[next]=id2;
		LSP[id1].cost[next]=cost;

		next=LSP[id2].numberOfLinks++;
		if (next>=MAX_LINKS) error("max number of links per router exceeded");
		LSP[id2].neighborID[next]=id1;
		LSP[id2].cost[next]=cost;
	}

	fclose(in);
}

// Generates a random network
// numberOfRouters	Input	Number Of Routers
// maxNumberOfLinks	Input	Maximum Number Of Links for each router
// LSP				Output	Pointer to a heap array of numberOfRouters LSPs
void generateRandomNetwork(unsigned int numberOfRouters, unsigned int maxNumberOfLinks, Packet* &LSP)
{
	unsigned int i;

	// prepare HELLO packets
	LSP=new Packet[numberOfRouters];
	for (i=0 ; i<numberOfRouters ; i++)
	{
		LSP[i].type=HELLO;
		LSP[i].sourceID=i;
		LSP[i].numberOfLinks=0;
	}

	unsigned int id1, id2, cost, next;
	for (id1=0 ; id1<numberOfRouters ; id1++) // this will prevent double links
	{
		id2=id1;
		for (i=0 ; i<maxNumberOfLinks ; i++)
		{
			id2+=1+rand()*(numberOfRouters-1)/RAND_MAX/maxNumberOfLinks*2; // incremental
			if (id2>=numberOfRouters) break;
			cost=rand()*10/RAND_MAX;
			printf("%u\t%u\t%u\n", id1, id2, cost);

			if (id1>=numberOfRouters || id2>=numberOfRouters) error("router's ID must be < number of routers");

			next=LSP[id1].numberOfLinks++;
			if (next>=MAX_LINKS) error("max number of links per router exceeded");
			LSP[id1].neighborID[next]=id2;
			LSP[id1].cost[next]=cost;

			next=LSP[id2].numberOfLinks++;
			if (next>=MAX_LINKS) error("max number of links per router exceeded");
			LSP[id2].neighborID[next]=id1;
			LSP[id2].cost[next]=cost;
		}
	}
}

// Prints the graph computed by a router (for debug purposes)
void saveLocalNet(unsigned int myID, Packet* *LSP, unsigned int n)
{
	FILE* out;
	char filename[18];
	unsigned int i, j;
	
	sprintf(filename, "net%05u.txt", myID);
	out=fopen(filename, "wt");
	if (out==0) error("cannot open output file");

	fprintf(out, "Nodes: %u\n", n);
	for (i=0 ; i<n ; i++) if (LSP[i]) // skip unreachables
	{
		fprintf(out, "-------------------\nNeighbors of: %u\n-------------------\n", LSP[i]->sourceID);
		for (j=0 ; j<LSP[i]->numberOfLinks ; j++)
		{
			fprintf(out, "%u\t%u\n", LSP[i]->neighborID[j], LSP[i]->cost[j]);		
		}
	}

	fclose(out);
}

// Prints the shortest path to every other Router
void saveRoutes(unsigned int myID, Dijkstra_label *label, unsigned int n)
{
	FILE* out;
	char filename[18];
	unsigned int path[80];
	unsigned int i, j, k;
	
	sprintf(filename, "out%05u.txt", myID);
	out=fopen(filename, "wt");
	if (out==0) error("cannot open output file");

	fprintf(out, "-------------------------------------------------------------------------------\n");
	fprintf(out, " Destination    Distance      Path - ID(Cost)\n");
	fprintf(out, "-------------------------------------------------------------------------------\n");
	for (i=0 ; i<n ; i++) if (label[i].predecessor!=-1) // skip unreachables (and the source)
	{
		fprintf(out, "%12u%12u      ", i, label[i].distance);

		// backtrace the path from i to myID...
		k=0;
		for (j=i ; j!=myID ; j=label[j].predecessor)
		{
			path[k]=j;
			k++;
			if (k>=80) k=0;
		}
		// ...and print it in reversed order
		while (k--) fprintf(out, "-> %u(%u) ", path[k], label[path[k]].distance);
		fprintf(out, "\n");
	}
	fprintf(out, "-------------------------------------------------------------------------------\n");

	fclose(out);
}

// Traverses the graph and determines if the adjacency list is complete
// Check if all reachables have completed too
bool graphCompleted(unsigned int myID, Packet* *LSP, unsigned int n)
{
	bool done;
	unsigned int i, j;

	// If myID has completed...
	for (i=0 ; i<n ; i++)
		if (LSP[i])	for (j=0 ; j<LSP[i]->numberOfLinks ; j++) if (LSP[LSP[i]->neighborID[j]]==NULL) return FALSE;
	g_completed[myID]=TRUE;

	// ...and all reachables have completed
	done=TRUE;
	for (i=0 ; i<n ; i++)
		if (LSP[i]) done = done && g_completed[i];
	return done;
}

//-------------------------------------------------------------------------
// Client and Server Sockets on localhost
//-------------------------------------------------------------------------

// returns a socket listening on localhost:port
int listenOnPort(int port)
{
	int s, code;

	s=socket(AF_INET, SOCK_STREAM, 0);
	if (s<0) error("socket");

	const char yes='1';
	code=setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)); // reuse port
	if (code<0) error("setsockopt");

	struct sockaddr_in address;
	address.sin_family=AF_INET;
	address.sin_port=htons(port);
	address.sin_addr.s_addr=inet_addr("127.0.0.1"); // loopback
	memset(&address.sin_zero, 0, sizeof(address.sin_zero));

	code=bind(s, (struct sockaddr *)&address, sizeof(struct sockaddr));
	if (code<0) error("bind");

	code=listen(s, QUEUE_SIZE);
	if (code<0) error("listen");

	return s;
}

// returns a socket connected to localhost:port
int connectToPort(int port)
{
	int s, code;

	s=socket(PF_INET, SOCK_STREAM, 0);
	if (s<0) error("socket");

	struct sockaddr_in address;
	address.sin_family= AF_INET;
	address.sin_port= htons(port);
	address.sin_addr.s_addr=inet_addr("127.0.0.1"); // loopback
	memset(&address.sin_zero, 0, sizeof(address.sin_zero));

	do {
		code=connect(s, (struct sockaddr *)&address, sizeof(struct sockaddr));
	} while (code<0 && WSAGetLastError()==WSAECONNREFUSED); // retry if queue full
	if (code<0) error("connect");

	return s;
}

//-------------------------------------------------------------------------
// Slave
//-------------------------------------------------------------------------

// Simulates the network-layer of a router (spawned in a Thread)
// Virtual Connections between routers are simulated by Sockets
//
// Link State Routing steps simulated here:
// 4) Send LSP to all other routers (FLOODING)
// 5) Compute the shortest path to every other router (DIJKSTRA)
// 6) Update routing table (FORWARDING)
int Router(int *lpdwParam) { 
	int sMaster;
	Packet myLSP, *incoming;
	unsigned int &myID=myLSP.sourceID;
	unsigned int i, j;

	sMaster=connectToPort(MASTER_PORT); // connect to Master

	// Receive an HELLO packet telling who my neighbors are (from the Master)
	recv(sMaster, (char*)&myLSP, sizeof(Packet), 0);
	if (myLSP.type!=HELLO) error("expected HELLO");

	int *sLink=new int[myLSP.numberOfLinks];

	// Connect a socket to each neighbor
	// For every link, the router with higher ID will do the Server and the other the Client
	// This hierarchical agreement will prevent deadlock (i.e. n routers, linked in a ring, waiting for each other)
	int sListener=listenOnPort(BASE_PORT+myLSP.sourceID);
	for (i=0 ; i<myLSP.numberOfLinks ; i++) 
	{
		if (myID>myLSP.neighborID[i]) sLink[i]=accept(sListener, NULL, NULL); // It can accept in the wrong order: never mind
		else sLink[i]=connectToPort(BASE_PORT+myLSP.neighborID[i]);
	}
	closesocket(sListener);

	// Send HACK to Master
	// TODO: send port numbers as required (but why?)
	incoming=new Packet;
	incoming->type=HACK;
	incoming->sourceID=myID;
	send(sMaster, (char*)incoming, sizeof(Packet), 0);

	// Wait START form Master
	recv(sMaster, (char*)incoming, sizeof(Packet), 0);
	if (incoming->type!=START) error("expected START");
	delete incoming;
	printf("FLOODING: Router %d START\n", myID);

	//-------------------------------------------------------------------------
	// 4) Send LSP to all other routers (FLOODING)
	//-------------------------------------------------------------------------

	// Now the simulated network is online:
	//   this thread <--> network layer of this router
	// send and recv <--> interface with the Data Link layer (HP: connection-oriented service)
	//      sLink[i] <--> handle for a data link connection (with a neighbor router)
	//        socket <--> network-layer-virtual-connection between two neighbors routers
	
	// Link State database
	Packet* LSP[MAX_ROUTERS]; // will maintain the adjacency list representation of the network graph
	for (i=0 ; i<MAX_ROUTERS ; i++) LSP[i]=NULL;
	bool* ACKflags[MAX_ROUTERS]; // will maintain the ACK flags waiting-list for each LSP flooded
	for (i=0 ; i<MAX_ROUTERS ; i++) ACKflags[i]=NULL;
	bool lostLSPs;
	unsigned int maxIDSeen;

	// Prepare for select()
	fd_set allLinks;
	fd_set readyLinks;
	FD_ZERO(&allLinks);
	for (i=0 ; i<myLSP.numberOfLinks ; i++) FD_SET(sLink[i], &allLinks);
	struct timeval timeout; // TODO: check if some implementations of select() can change this
	timeout.tv_sec=FLOODING_TIMEOUT;
	timeout.tv_usec=0;

	// Send local LSP to all the neighbors
	myLSP.type=LSU;
	for (i=0 ; i<myLSP.numberOfLinks ; i++) send(sLink[i], (char*)&myLSP, sizeof(Packet), 0); // TODO: can I rely on buffers?
	// Update the LS database
	LSP[myID]=&myLSP;
	ACKflags[myID]=new bool[myLSP.numberOfLinks]; // add ACK flags for myLSP
	for (i=0 ; i<myLSP.numberOfLinks ; i++) ACKflags[myID][i]=TRUE; // expect an ACK from every link
	maxIDSeen=myID;

	do {
		// Store and then bounce incoming LSPs to other data links
		// (but only if they are new: this will dampen the flooding)
		readyLinks=allLinks;
		while (select(0, &readyLinks, NULL, NULL, &timeout)>0) // wait for incoming packets (select() returns 0 on timeout)
		{
			for (i=0 ; i<myLSP.numberOfLinks ; i++) if (FD_ISSET(sLink[i], &readyLinks)) 
			{
				// Incoming packet from link i; let's get it
				incoming=new Packet;
				recv(sLink[i], (char*)incoming, sizeof(Packet), 0);

				if (rand()<RAND_MAX/20) incoming->type=HACK; // Simulates 5% packet loss

				switch (incoming->type)
				{
				case LSU:
					// LSP from link i

					// Send an ACK on link i
					incoming->type=ACK;
					send(sLink[i], (char*)incoming, sizeof(Packet), 0);
					incoming->type=LSU;

					// Check if it is new
					if (LSP[incoming->sourceID]==NULL)
					{
						// Flood it...
						for (j=0 ; j<myLSP.numberOfLinks ; j++)
							if(j!=i) // source neighbor already has it
								send(sLink[j], (char*)incoming, sizeof(Packet), 0);

						// ...and Update the LS database
						LSP[incoming->sourceID]=incoming; // store it
						ACKflags[incoming->sourceID]=new bool[myLSP.numberOfLinks]; // add ACK flags for this LSP
						for (j=0 ; j<myLSP.numberOfLinks ; j++) ACKflags[incoming->sourceID][j]=TRUE; // expect an ACK from every link...
						ACKflags[incoming->sourceID][i]=FALSE; // ...source link excluded
						if (incoming->sourceID>maxIDSeen) maxIDSeen=incoming->sourceID;

						printf("FLOODING: Router %u has got the LSP of %u\n", myID, incoming->sourceID);
					}
					else delete incoming; // already seen ==> discard it		
					break;

				case ACK:
					// ACK from link i, for the LSP of 'incoming->sourceID'
					ACKflags[incoming->sourceID][i]=FALSE;
					//printf("ACK: Router %u has got ACK %u from link %u\n", myID, incoming->sourceID, i);
					delete incoming;
					break;

				case DATA:
					// Forward it according to its destination and current Routing Table
					// TODO: (not required)
					delete incoming;
					break;

				default:
					delete incoming;
					break;
				}
			}		
			readyLinks=allLinks;
		}

		// Timeout ==> Flooding (should be) completed; to be sure...

		// ...check the ACK flags and resend lost LSPs
		lostLSPs=FALSE;
		for (i=0 ; i<=maxIDSeen ; i++)
			if (ACKflags[i])
				for (j=0 ; j<myLSP.numberOfLinks ; j++)
					if (ACKflags[i][j])
					{
						lostLSPs=TRUE;
						send(sLink[j], (char*)LSP[i], sizeof(Packet), 0);
						printf("ACK: Router %u had to resend LSP %u on link %u\n", myID, i, j);
					}

	} while (lostLSPs || !graphCompleted(myID, LSP, maxIDSeen+1));

	printf("FLOODING: Router %d DONE\n", myID);
	unsigned int n=maxIDSeen+1; // number of Routers (if the network is a connected graph)
	//saveLocalNet(myID, LSP, n);

	//-------------------------------------------------------------------------
	// 5) Compute the shortest path to every other router (DIJKSTRA)
	//-------------------------------------------------------------------------

	// This is the simplest form of Dijkstra: O(n^2)
	// It fits perfectly our adjacency list and requires less memory accesses
	// TODO: Try with a Priority queue and a Fibonacci heap

	unsigned int k, min, distance_via_k;
	Dijkstra_label *label=new Dijkstra_label[n];
	Dijkstra_label *neighborLabel;

	// Init labels
	for (i=0 ; i<n ; i++)
	{
		label[i].predecessor=-1;
		label[i].distance=-1; // max unsigned int
		label[i].permanent=FALSE;
	}
	label[myID].distance=0;

	k=myID; // k is the current working node
	do
	{
		label[k].permanent=TRUE;

		// Relax the links of the working node
		for (i=0 ; i<LSP[k]->numberOfLinks ; i++)
		{
			neighborLabel=&label[LSP[k]->neighborID[i]];
			distance_via_k=label[k].distance+LSP[k]->cost[i];
			if (!neighborLabel->permanent && distance_via_k<neighborLabel->distance)
			{
				neighborLabel->predecessor=k;
				neighborLabel->distance=distance_via_k;
			}
		}

		// Find the tentative node with the smallest distance
		min=-1; // max unsigned int
		for (i=0 ; i<n ; i++)
			if (!label[i].permanent && label[i].distance<min)
			{
				min=label[i].distance;
				k=i;
			}
	} while (label[k].permanent==FALSE);

	//-------------------------------------------------------------------------
	// 6) Update routing table (FORWARDING)
	//-------------------------------------------------------------------------
	
	// Print the best route to each destination (the routing table is part of it)
	saveRoutes(myID, label, n);
	
	// The End: indeed, the network layer of a real router simply loops forever (daemon)
	delete[] label;
	for (i=0 ; i<n ; i++) if (LSP[i])
	{
		if (i!=myID) delete LSP[i];
		delete[] ACKflags[i];
	}
	//select(0, &allLinks, NULL, NULL, &timeout); // must wait for the others, before closing connections
	//for (i=0 ; i<myLSP.numberOfLinks ; i++) closesocket(sLink[i]);
	//delete[] sLink;
	closesocket(sMaster);
    return 0; 
} 

//-------------------------------------------------------------------------
// Master
//-------------------------------------------------------------------------

// Spawns Routers and uploads neighbor-delay data to them (via sockets)
//
// Link State Routing steps assolved here (for each router):
// 1) Find neighbors and their address (HELLO)
// 2) Measure delay to them (ECHO)
// 3) Build Link State Packet (LSP)
void Master()
{
	unsigned int i;
	int sListener, *sSlave; // sockets
	int bytes;

	unsigned int n=0; // Number Of Routers
	Packet *LSP=NULL; // Link State Packets

	printf("Loading network...\n");
	loadNetworkDescription(n, LSP);

	sSlave=new int[n]; // sockets connected to Slaves

	sListener=listenOnPort(MASTER_PORT); // setup listening-socket

	// Create a new thread for each Router
	HANDLE *hSlave=new HANDLE[n]; // used later to wait their end
	DWORD dwThreadId;
    for (i=0 ; i<n ; i++) hSlave[i]=CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)Router, NULL, 0, &dwThreadId);

	// Wait connections from Slaves
	for (i=0 ; i<n ; i++)
	{
		sSlave[i]=accept(sListener, NULL, NULL);
		if (sSlave[i]<0) error("accept");
	}
	closesocket(sListener);

	// Upload LSPs to Routers
	for (i=0 ; i<n ; i++)
	{
		bytes=send(sSlave[i], (char*)&LSP[i], sizeof(LSP[i]), 0);
		if (bytes!=sizeof(LSP[i])) error("LSP fragmented"); // never happens, in this case...
	}
	delete[] LSP;

	// Wait HACK from Routers
	Packet *buf=new Packet;
	for (i=0 ; i<n ; i++)
	{
		recv(sSlave[i], (char*)buf, sizeof(Packet), 0);
		if (buf->type!=HACK) error("expected HACK");
	}

	printf("Network up and running (try netstat)\n\n");
	printf("5%% packet loss will be simulated, with a %us timeout for ACKs\n", FLOODING_TIMEOUT);
	printf("Press any key to start Flooding...\n");
	getc(stdin);
	g_completed=new bool[n];
	for (i=0 ; i<n ; i++) g_completed[i]=FALSE;
	
	// Our simulated network is up and running: time to start the Link State Routing
	// Send a START packet to each Router
	buf->type=START;
	for (i=0 ; i<n ; i++) send(sSlave[i], (char*)buf, sizeof(Packet), 0);
	delete buf;

	// Wait the end of all the threads
	for (i=0 ; i<n ; i++) WaitForSingleObject(hSlave[i], INFINITE);
	delete[] hSlave;

	// Close connections with Slaves
	for (i=0 ; i<n ; i++) closesocket(sSlave[i]);
	delete[] sSlave;

	delete[] g_completed;
}

//-------------------------------------------------------------------------
// main
//-------------------------------------------------------------------------

int main(int argc, char* argv[])
{
	printf("LinkStateSim 1.3 by Marco Benvegnu\'\nLink State Routing simulator with Sockets\n\n");
	if (argc>1) g_inputPath=argv[1]; // override default input filename

	// StartUp Sockets in Win32
	WSADATA ws;
	if (WSAStartup(MAKEWORD(1,1), &ws)) error("failed to find WinSock 1.1 or better"); 

	/*
	// just for fun
	char buf[80];
	gethostname(buf, sizeof(buf));
	printf("%s ",buf);

	struct hostent *h;
	h=gethostbyname(buf);
	printf("IP Address : %s\n", inet_ntoa(*((struct in_addr *)h->h_addr)));
	*/

	Master();

	printf("\nSimulation completed. Press any key to continue...\n");
	getc(stdin);

	WSACleanup();
	return 0;
}

